@@ -10,7 +10,9 @@ __all__ = [
 ]
 
 from time import time
-from socket import GNUTLS_SHUT_RDWR as SOCKET_SHUT_RDWR
+from socket import SHUT_RDWR
+
+from typing import Optional
 
 from _ctypes import PyObj_FromPtr
 from ctypes import (
@@ -302,7 +304,7 @@ class Session(object):
     """Abstract class representing a TLS session created from a TCP socket
        and a Credentials object."""
 
-    session_type = (
+    session_type: Optional[int] = (
         None  # placeholder for GNUTLS_SERVER or GNUTLS_CLIENT as defined by subclass
     )
 
@@ -431,7 +433,7 @@ class Session(object):
     def bye(self, how=GNUTLS_SHUT_RDWR):
         gnutls_bye(self._c_object, how)
 
-    def shutdown(self, how=SOCKET_SHUT_RDWR):
+    def shutdown(self, how=SHUT_RDWR):
         self.socket.shutdown(how)
 
     def close(self):
@@ -534,7 +536,7 @@ class ServerSessionFactory(object):
         session = self.session_class(new_sock, self.context)
         return (session, address)
 
-    def shutdown(self, how=SOCKET_SHUT_RDWR):
+    def shutdown(self, how=SHUT_RDWR):
         self.socket.shutdown(how)
 
     def close(self):
